from sqlalchemy import text

__all__ = ["recreate_views"]


def recreate_views(db, file_version, all_views, views_to_delete):
    """Recreate predefined views in a ThreediDatabase instance"""
    engine = db.get_engine()

    with engine.begin() as connection:
        for name, view in all_views.items():
            connection.execute(text(f"DROP VIEW IF EXISTS {name}"))
            connection.execute(
                text(f"DELETE FROM views_geometry_columns WHERE view_name = '{name}'")
            )
            connection.execute(text(f"CREATE VIEW {name} AS {view['definition']}"))
            if file_version == 3:
                connection.execute(
                    text(
                        f"INSERT INTO views_geometry_columns (view_name, view_geometry,view_rowid,f_table_name,f_geometry_column) VALUES('{name}', '{view['view_geometry']}', '{view['view_rowid']}', '{view['f_table_name']}', '{view['f_geometry_column']}')"
                    )
                )
            else:
                connection.execute(
                    text(
                        f"INSERT INTO views_geometry_columns (view_name, view_geometry,view_rowid,f_table_name,f_geometry_column,read_only) VALUES('{name}', '{view['view_geometry']}', '{view['view_rowid']}', '{view['f_table_name']}', '{view['f_geometry_column']}', 0)"
                    )
                )
        for name in views_to_delete:
            connection.execute(text(f"DROP VIEW IF EXISTS {name}"))
            connection.execute(
                text(f"DELETE FROM views_geometry_columns WHERE view_name = '{name}'")
            )



from threedi_schema.domain import models


def test_read_geometry(gpkg_db):
    channel = gpkg_db.get_session().query(models.Culvert).first()
    assert channel.code == "40902275"
    assert channel.the_geom is not None  # this is what happens with GeoAlchemy2



from geoalchemy2.types import Geometry
from sqlalchemy import func, text

__all__ = ["ensure_spatial_indexes"]


def _ensure_spatial_index(connection, column):
    """Ensure presence of a spatial index for given geometry olumn"""
    if (
        connection.execute(
            func.RecoverSpatialIndex(column.table.name, column.name)
        ).scalar()
        is not None
    ):
        return False

    idx_name = f"{column.table.name}_{column.name}"
    connection.execute(text(f"DROP TABLE IF EXISTS idx_{idx_name}"))
    for prefix in {"gii_", "giu_", "gid_"}:
        connection.execute(text(f"DROP TRIGGER IF EXISTS {prefix}{idx_name}"))
    if (
        connection.execute(
            func.CreateSpatialIndex(column.table.name, column.name)
        ).scalar()
        != 1
    ):
        raise RuntimeError(f"Spatial index creation for {idx_name} failed")

    return True


def ensure_spatial_indexes(db, models):
    """Ensure presence of spatial indexes for all geometry columns"""
    created = False
    engine = db.get_engine()

    with engine.begin() as connection:
        for model in models:
            geom_columns = [
                x for x in model.__table__.columns if isinstance(x.type, Geometry)
            ]
            if len(geom_columns) > 1:
                # Pragmatic fix: spatialindex breaks on multiple geometry columns per table
                geom_columns = [x for x in geom_columns if x.name == "the_geom"]
            if geom_columns:
                created &= _ensure_spatial_index(connection, geom_columns[0])

        if created:
            connection.execute(text("VACUUM"))



import pathlib

from setuptools import find_packages, setup

long_description = "\n\n".join([open("README.rst").read(), open("CHANGES.rst").read()])


def get_version():
    # Edited from https://packaging.python.org/guides/single-sourcing-package-version/
    init_path = pathlib.Path(__file__).parent / "threedi_schema/__init__.py"
    for line in init_path.open("r").readlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


install_requires = [
    "GeoAlchemy2>=0.9,!=0.11.*",
    "SQLAlchemy>=1.4",
    "alembic>=1.8,<2",
]

tests_require = ["pytest", "pytest-cov"]
cli_require = ["click"]


setup(
    name="threedi-schema",
    version=get_version(),
    description="The schema of 3Di schematization files",
    long_description=long_description,
    # Get strings from http://www.python.org/pypi?%3Aaction=list_classifiers
    classifiers=["Programming Language :: Python", "Framework :: Django"],
    keywords=[],
    author="Nelen & Schuurmans",
    author_email="info@nelen-schuurmans.nl",
    url="https://github.com/nens/threedi-schema",
    license="MIT",
    packages=find_packages(include="threedi_schema*"),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    python_requires=">=3.7",
    tests_require=tests_require,
    extras_require={"test": tests_require, "cli": cli_require},
    entry_points={
        "console_scripts": [
            "threedi_schema = threedi_schema.scripts:main"
        ]
    },
)



from sqlalchemy import func

from threedi_schema.domain import models
from threedi_schema.infrastructure.spatialite_versions import (
    copy_model,
    get_spatialite_version,
)


def test_get_spatialite_version(empty_sqlite_v3):
    lib_version, file_version = get_spatialite_version(empty_sqlite_v3)
    assert lib_version in (3, 4, 5)
    assert file_version == 3


def test_copy_model(empty_sqlite_v3, empty_sqlite_v4):
    db_from = empty_sqlite_v3
    db_to = empty_sqlite_v4

    # Add a record to 'db_from'
    obj = models.ConnectionNode(
        id=3, code="test", the_geom="SRID=4326;POINT(-71.064544 42.287870)"
    )
    with db_from.session_scope() as session:
        session.add(obj)
        session.commit()

    # Copy it
    copy_model(db_from, db_to, models.ConnectionNode)

    # Check if it is present in 'db_to'
    with db_to.session_scope() as session:
        records = list(
            session.query(
                models.ConnectionNode.id,
                models.ConnectionNode.code,
                func.ST_AsText(models.ConnectionNode.the_geom),
                models.ConnectionNode.the_geom_linestring,
            )
        )

        assert records == [(3, "test", "POINT(-71.064544 42.28787)", None)]


def test_copy_invalid_geometry(empty_sqlite_v3, empty_sqlite_v4):
    """Copying an invalid geometry (ST_IsValid evaluates to False) is possible"""
    db_from = empty_sqlite_v3
    db_to = empty_sqlite_v4

    obj = models.Surface(
        id=3,
        code="test",
        display_name="test",
        the_geom="SRID=4326;POLYGON((0 0, 10 10, 0 10, 10 0, 0 0))",
        surface_parameters_id=1,
    )
    with db_from.session_scope() as session:
        session.add(obj)
        session.commit()

    copy_model(db_from, db_to, models.Surface)

    with db_to.session_scope() as session:
        records = list(
            session.query(
                models.Surface.id,
                func.ST_AsText(models.Surface.the_geom),
            )
        )

        assert records == [(3, "POLYGON((0 0, 10 10, 0 10, 10 0, 0 0))")]



"""Removed unused v1 tables.

Revision ID: 0202
Revises:
Create Date: 2021-09-29 13:50:19.544275

"""
import re

from alembic import op
from sqlalchemy import inspect, text

# revision identifiers, used by Alembic.
revision = "0202"
down_revision = "0201"
branch_labels = None
depends_on = None


V1_TABLES = [
    "boundary_line",
    "boundary_line_chk",
    "boundary_point",
    "boundary_point_chk",
    "bridge",
    "bridge_chk",
    "channel",
    "channel_chk",
    "check_boundary_nodes",
    "cross_section",
    "cross_section_chk",
    "cross_section_definition",
    "cross_section_definition_chk",
    "culvert",
    "culvert_chk",
    "floodfill",
    "global_settings",
    "grid_refinement",
    "initial_waterlevel",
    "initial_waterlevel_chk",
    "lateral",
    "lateral_chk",
    "levee",
    "levee_chk",
    "manhole_2d",
    "manhole_2d_chk",
    "orifice",
    "orifice_chk",
    "potential_cross_section",
    "pumped_drainage_area",
    "pumped_drainage_area_chk",
    "pumpstation",
    "pumpstation_chk",
    "qualitycheckresult",
    "rain_event",
    "settings",
    "sewerage_boundary_conditions",
    "sewerage_boundary_conditions_chk",
    "sewerage_cross_section_definition",
    "sewerage_cross_section_definition_chk",
    "sewerage_impervious_surface",
    "sewerage_impervious_surface_chk",
    "sewerage_impervious_surface",
    "sewerage_impervious_surface_pipe_map",
    "sewerage_lateral",
    "sewerage_lateral_chk",
    "sewerage_manhole",
    "sewerage_manhole_chk",
    "sewerage_orifice",
    "sewerage_orifice_chk",
    "sewerage_outlet",
    "sewerage_outlet_chk",
    "sewerage_pipe",
    "sewerage_pipe_chk",
    "sewerage_pumpstation",
    "sewerage_pumpstation_chk",
    "sewerage_weir",
    "sewerage_weir_chk",
    "sisp_map_chk",
    "weir",
    "weir_chk",
    "wind",
]

# Don't try to delete Spatialite views.
# See https://www.gaia-gis.it/fossil/libspatialite/wiki?name=GetDbObjectScope%28%29
SPATIALITE_VIEWS = [
    "geom_cols_ref_sys",
    "spatial_ref_sys_all",
    "vector_coverages_ref_sys",
]

options = "|".join(V1_TABLES)
VIEW_REGEX = re.compile(f".*(\\w+{options}\\w+).*")


def upgrade():
    # first list views that refer to V1 TABLES
    conn = op.get_bind()
    inspector = inspect(conn)
    view_names = inspector.get_view_names()
    for view_name in view_names:
        if view_name.startswith("v2_") or view_name in SPATIALITE_VIEWS:
            continue
        defn = inspector.get_view_definition(view_name)
        if VIEW_REGEX.match(defn):
            op.execute(text(f"DROP VIEW {view_name}"))

    # then delete the actual tables if they exist
    existing = set(inspector.get_table_names())
    to_delete = set(existing).intersection(V1_TABLES)
    for table_name in to_delete:
        op.drop_table(table_name)


def downgrade():
    pass



"""clean the tables associated with 1D-2D exchange

Revision ID: 0214
Revises: 0213
Create Date: 2022-12-22 11:42:00

"""
import sqlalchemy as sa
from alembic import op

from threedi_schema.domain.custom_types import Geometry

# revision identifiers, used by Alembic.
revision = "0214"
down_revision = "0213"
branch_labels = None
depends_on = None


LEVEE_TO_OBSTACLE = """
INSERT INTO v2_obstacle (code, crest_level, the_geom) SELECT code, crest_level, the_geom FROM v2_levee
"""

OBSTACLE_TO_LEVEE = """
INSERT INTO v2_levee (code, crest_level, the_geom) SELECT code, crest_level, the_geom FROM v2_obstacle
"""


def upgrade():
    op.execute(sa.text(LEVEE_TO_OBSTACLE))

    for table_name in ["v2_connected_pnt", "v2_calculation_point", "v2_levee"]:
        op.execute(sa.text(f"SELECT DropGeoTable('{table_name}')"))


def downgrade():
    op.create_table(
        "v2_levee",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("code", sa.String(length=100), nullable=True),
        sa.Column("crest_level", sa.Float(), nullable=True),
        sa.Column("the_geom", Geometry("LINESTRING"), nullable=True),
        sa.Column("material", sa.Integer(), nullable=True),
        sa.Column("max_breach_depth", sa.Float(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "v2_calculation_point",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("content_type_id", sa.Integer(), nullable=True),
        sa.Column("user_ref", sa.String(length=80), nullable=True),
        sa.Column("calc_type", sa.Integer(), nullable=True),
        sa.Column("the_geom", Geometry("POINT"), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_table(
        "v2_connected_pnt",
        sa.Column("id", sa.Integer(), nullable=False),
        sa.Column("calculation_pnt_id", sa.Integer(), nullable=True),
        sa.Column("levee_id", sa.Integer(), nullable=True),
        sa.Column("exchange_level", sa.Float(), nullable=True),
        sa.Column("the_geom", Geometry("POINT"), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )



"""clean unused connected points and transform remaining to potential breaches

Revision ID: 0213
Revises: 0212
Create Date: 2022-12-21 14:54:00

"""
import logging

from alembic import op
from sqlalchemy import Column, Float, ForeignKey, func, Integer, String
from sqlalchemy.orm import declarative_base, Session

from threedi_schema.domain.custom_types import Geometry

# revision identifiers, used by Alembic.
revision = "0213"
down_revision = "0212"
branch_labels = None
depends_on = None

logger = logging.getLogger(__name__)

## Copy of the ORM at this point in time:

Base = declarative_base()


class Levee(Base):
    __tablename__ = "v2_levee"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    crest_level = Column(Float)
    the_geom = Column(
        Geometry(
            "LINESTRING"
        ),
        nullable=False
    )
    material = Column(Integer)
    max_breach_depth = Column(Float)


class CalculationPoint(Base):
    __tablename__ = "v2_calculation_point"
    id = Column(Integer, primary_key=True)

    content_type_id = Column(Integer)
    user_ref = Column(String(80), nullable=False)
    calc_type = Column(Integer)
    the_geom = Column(
        Geometry("POINT"),
        nullable=False
    )


class ConnectedPoint(Base):
    __tablename__ = "v2_connected_pnt"
    id = Column(Integer, primary_key=True)

    calculation_pnt_id = Column(
        Integer, ForeignKey(CalculationPoint.__tablename__ + ".id"), nullable=False
    )
    levee_id = Column(Integer, ForeignKey(Levee.__tablename__ + ".id"))

    exchange_level = Column(Float)
    the_geom = Column(
        Geometry("POINT"),
        nullable=False
    )


class PotentialBreach(Base):
    __tablename__ = "v2_potential_breach"
    id = Column(Integer, primary_key=True)
    code = Column(String(100))
    display_name = Column(String(255))
    exchange_level = Column(Float)
    maximum_breach_depth = Column(Float)
    levee_material = Column(Integer)
    the_geom = Column(
        Geometry("LINESTRING", from_text="ST_GeomFromEWKB"),
        nullable=False
    )
    channel_id = Column(Integer, nullable=False)


class ConnectionNode(Base):
    __tablename__ = "v2_connection_nodes"
    id = Column(Integer, primary_key=True)
    the_geom = Column(
        Geometry("POINT"),
        nullable=False
    )


class Manhole(Base):
    __tablename__ = "v2_manhole"

    id = Column(Integer, primary_key=True)
    connection_node_id = Column(
        Integer,
        ForeignKey(ConnectionNode.__tablename__ + ".id"),
        nullable=False,
        unique=True,
    )


class Channel(Base):
    __tablename__ = "v2_channel"

    id = Column(Integer, primary_key=True)
    calculation_type = Column(Integer, nullable=False)
    dist_calc_points = Column(Float, nullable=True)
    connection_node_start_id = Column(Integer, nullable=False)
    connection_node_end_id = Column(Integer, nullable=False)
    the_geom = Column(
        Geometry(
            "LINESTRING" 
        ),
        nullable=True
    )


def parse_connected_point_user_ref(user_ref: str):
    """Return content_type, content_id, node_number from a user_ref.

    Raises Exception for various parse errors.

    Example
    -------
    >>> parse_connected_point_user_ref("201#123#v2_channels#4)
    ContentType.TYPE_V2_CHANNEL, 123, 3
    """
    _, id_str, type_str, node_nr = user_ref.split("#")
    return type_str, int(id_str), int(node_nr)


def clean_connected_points(session):
    conn_point_ids = [
        x[0]
        for x in session.query(ConnectedPoint.id)
        .join(CalculationPoint, isouter=True)
        .filter(
            (ConnectedPoint.the_geom != CalculationPoint.the_geom)
            | (ConnectedPoint.exchange_level > -9999.0)
            | (ConnectedPoint.levee_id != None)
        )
        .all()
    ]
    session.query(ConnectedPoint).filter(
        ConnectedPoint.id.notin_(conn_point_ids)
    ).delete(synchronize_session="fetch")
    calc_point_ids = [
        x[0]
        for x in session.query(ConnectedPoint.calculation_pnt_id)
        .filter(ConnectedPoint.id.in_(conn_point_ids))
        .all()
    ]
    session.query(CalculationPoint).filter(
        CalculationPoint.id.notin_(calc_point_ids)
    ).delete(synchronize_session="fetch")
    return conn_point_ids


def get_channel_id(session, user_ref):
    """Get channel id and index into channel nodes"""
    type_ref, pk, node_nr = parse_connected_point_user_ref(user_ref)
    if type_ref == "v2_channel":
        return pk, node_nr - 1
    elif type_ref == "v2_manhole":
        return get_channel_id_manhole(session, pk)
    return None, None


def get_channel_id_manhole(session, pk):
    obj = session.query(Manhole).filter(Manhole.id == pk).first()
    if obj is None:
        return None, None
    connection_node_id = obj.connection_node_id

    # find a channel connected to this connection node, with connected calc type
    channels = (
        session.query(Channel)
        .filter(
            (
                (Channel.connection_node_start_id == connection_node_id)
                | (Channel.connection_node_end_id == connection_node_id)
            )
            & Channel.calculation_type.in_([102, 105])
        )
        .all()
    )
    if len(channels) == 0:
        return None, None

    # prefer double connected, and then prefer lowest id
    channel = sorted(channels, key=lambda x: (-x.calculation_type, x.id))[0]
    if channel.connection_node_start_id == connection_node_id:
        node_idx = 0
    else:
        node_idx = -1
    return channel.id, node_idx


def scalar_subquery(query):
    # compatibility between sqlalchemy 1.3 and 1.4
    try:
        return query.scalar_subquery()
    except AttributeError:
        return query.as_scalar()


def get_breach_line_geom(session, conn_point_id, channel_id, node_idx):
    channel = scalar_subquery(
        session.query(Channel.the_geom).filter(Channel.id == channel_id)
    )
    if node_idx == 0:
        breach_line_start = func.ST_PointN(channel, 1)
    elif node_idx == -1:
        breach_line_start = func.ST_PointN(channel, func.ST_NPoints(channel))
    else:
        breach_line_start = func.Snap(CalculationPoint.the_geom, channel, 1e-7)

    return (
        session.query(
            func.AsEWKB(func.MakeLine(breach_line_start, ConnectedPoint.the_geom))
        )
        .join(CalculationPoint)
        .filter(ConnectedPoint.id == conn_point_id)
        .one()[0]
    )


def to_potential_breach(session, conn_point_id):
    connected_point, calculation_point, levee = (
        session.query(
            ConnectedPoint,
            CalculationPoint,
            Levee,
        )
        .join(CalculationPoint)
        .join(Levee, isouter=True)
        .filter(ConnectedPoint.id == conn_point_id)
        .one()
    )

    channel_id, node_idx = get_channel_id(session, calculation_point.user_ref)
    if channel_id is None:
        return

    line_geom = get_breach_line_geom(session, conn_point_id, channel_id, node_idx)

    if connected_point.exchange_level not in (None, -9999.0):
        exchange_level = connected_point.exchange_level
    elif levee is not None:
        exchange_level = levee.crest_level
    else:
        exchange_level = None

    if levee is not None:
        maximum_breach_depth = levee.max_breach_depth
    else:
        maximum_breach_depth = None

    if exchange_level == -9999.0:
        exchange_level = None
    if maximum_breach_depth == -9999.0:
        maximum_breach_depth = None

    return PotentialBreach(
        code="#".join([str(connected_point.id), calculation_point.user_ref])[:100],
        exchange_level=exchange_level,
        maximum_breach_depth=maximum_breach_depth,
        levee_material=levee.material if levee is not None else None,
        the_geom=line_geom,
        channel_id=channel_id,
    )


def upgrade():
    session = Session(bind=op.get_bind())

    conn_point_ids = clean_connected_points(session)
    for conn_point_id in conn_point_ids:
        breach = to_potential_breach(session, conn_point_id)
        if breach is None:
            logger.warning(
                "Connected Point %d will be removed because it "
                "cannot be related to a channel. This may influence the "
                "1D-2D exchange of the model.",
                conn_point_id,
            )
        else:
            session.add(breach)
    session.flush()


def downgrade():
    pass



from .domain import constants, models

"""
Put features in beta development in these lists to prevent users from using them, in the SQLAlchemy format used in the schema.
"""
"""This list contains beta columns, e.g. models.SomeTable.some_column"""
BETA_COLUMNS = [
    models.GlobalSetting.vegetation_drag_settings_id,
    models.Manhole.exchange_thickness,
    models.Manhole.hydraulic_conductivity_in,
    models.Manhole.hydraulic_conductivity_out,
    models.Channel.exchange_thickness,
    models.Channel.hydraulic_conductivity_in,
    models.Channel.hydraulic_conductivity_out,
    models.Pipe.exchange_thickness,
    models.Pipe.hydraulic_conductivity_in,
    models.Pipe.hydraulic_conductivity_out,
]
"""
This list contains dicts with lists of beta values for columns, where the dict has the format:
{
    "columns": [
        models.SomeTable.some_column,
        models.SomeTable.some_other_column,
    ],
    "values": [
        constants.SomeConstantsClass.SOMEENUMVALUE,
        "or just some random string",
    ]
}
The modelchecker will go through each column and give an error when a beta value is used for its associated column.
"""
BETA_VALUES = [
    {
        "columns": [
            models.BoundaryConditions2D.boundary_type,  # 1d boundary conditions shouldn't have groundwater anyway
        ],
        "values": [
            constants.BoundaryType.GROUNDWATERLEVEL,
            constants.BoundaryType.GROUNDWATERDISCHARGE,
        ],
    }
]



"""remove surface_type

Revision ID: 0209
Revises: 0208
Create Date: 2022-10-13 10:45

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0209"
down_revision = "0208"
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("v2_surface_map") as batch_op:
        batch_op.drop_column("surface_type")


def downgrade():
    with op.batch_alter_table("v2_surface_map") as batch_op:
        batch_op.add_column(
            sa.Column("surface_type", sa.String(length=40), nullable=True)
        )
